/*
 * aTunes 1.11.2
 * Copyright (C) 2006-2009 Alex Aranda, Sylvain Gaudard, Thomas Beckers and contributors
 *
 * See http://www.atunes.org/wiki/index.php?title=Contributing for information about contributors
 *
 * http://www.atunes.org
 * http://sourceforge.net/projects/atunes
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

package net.sourceforge.atunes.kernel.controllers.playList;

import java.awt.Rectangle;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.File;
import java.util.ArrayList;
import java.util.List;

import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.Timer;

import net.sourceforge.atunes.gui.model.PlayListTableModel;
import net.sourceforge.atunes.gui.views.controls.playList.Column;
import net.sourceforge.atunes.gui.views.controls.playList.PlayListColumnClickedListener;
import net.sourceforge.atunes.gui.views.controls.playList.PlayListColumnSelector;
import net.sourceforge.atunes.gui.views.controls.playList.PlayListColumns;
import net.sourceforge.atunes.gui.views.controls.playList.PlayListTable;
import net.sourceforge.atunes.gui.views.panels.PlayListControlsPanel;
import net.sourceforge.atunes.gui.views.panels.PlayListPanel;
import net.sourceforge.atunes.kernel.Kernel;
import net.sourceforge.atunes.kernel.actions.Actions;
import net.sourceforge.atunes.kernel.actions.SavePlayListAction;
import net.sourceforge.atunes.kernel.actions.ShufflePlayListAction;
import net.sourceforge.atunes.kernel.controllers.model.PanelController;
import net.sourceforge.atunes.kernel.executors.BackgroundExecutor;
import net.sourceforge.atunes.kernel.handlers.DesktopHandler;
import net.sourceforge.atunes.kernel.handlers.PlayListHandler;
import net.sourceforge.atunes.kernel.handlers.PlayerHandler;
import net.sourceforge.atunes.kernel.handlers.VisualHandler;
import net.sourceforge.atunes.kernel.modules.repository.audio.AudioFile;
import net.sourceforge.atunes.kernel.modules.repository.tags.writer.TagEditionOperations;
import net.sourceforge.atunes.model.AudioObject;
import net.sourceforge.atunes.utils.log4j.LogCategories;

/**
 * The Class PlayListController.
 */
public class PlayListController extends PanelController<PlayListPanel> implements PlayListColumnClickedListener {

	/** The visible rect. */
	Rectangle visibleRect;

	/**
	 * Instantiates a new play list controller.
	 * 
	 * @param panel
	 *            the panel
	 */
	public PlayListController(PlayListPanel panel) {
		super(panel);
		addBindings();
		addStateBindings();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * net.sourceforge.atunes.kernel.controllers.model.Controller#addBindings()
	 */
	@Override
	protected void addBindings() {
		final PlayListTable table = getPanelControlled().getPlayListTable();
		table.addPlayListColumnClickedListener(this);

		// Set key listener for table
		table.addKeyListener(new PlayListKeyListener(this));

		PlayListListener listener = new PlayListListener(table, this);

		table.getArrangeColumns().addActionListener(listener);
		table.addMouseListener(listener);

		table.getSelectionModel().addListSelectionListener(listener);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * net.sourceforge.atunes.kernel.controllers.model.Controller#addStateBindings
	 * ()
	 */
	@Override
	protected void addStateBindings() {
		// Nothing to do
	}

	/**
	 * Arrange columns.
	 */
	public void arrangeColumns() {
		// Show column selector
		PlayListColumnSelector selector = VisualHandler.getInstance().getPlayListColumnSelector();
		selector.setColumns(PlayListColumns.getColumnsForSelection());
		selector.setVisible(true);

		// Apply changes
		((PlayListTableModel) getPanelControlled().getPlayListTable().getModel()).arrangeColumns(true);

	}

	/**
	 * Notifies when user clicks on a column header.
	 * 
	 * @param columnClicked
	 *            the column clicked
	 */
	public void columnClicked(Column columnClicked) {
		getLogger().debug(LogCategories.CONTROLLER, new String[] { columnClicked.getHeaderText() });

		// Sort play list
		PlayListHandler.getInstance().sortPlayList(columnClicked.getComparator());
	}

	/**
	 * Delete selection.
	 */
	public void deleteSelection() {
		getLogger().debug(LogCategories.CONTROLLER);

		PlayListTable table = VisualHandler.getInstance().getPlayListTable();
		int[] rows = table.getSelectedRows();
		if (rows.length > 0) {
			((PlayListTableModel) table.getModel()).removeSongs(rows);
			getPanelControlled().getPlayListTable().getSelectionModel().setSelectionInterval(-1, -1);
			PlayListHandler.getInstance().removeSongs(rows);
		}
	}

	/**
	 * Gets the main play list scroll pane.
	 * 
	 * @return the main play list scroll pane
	 */
	public JScrollPane getMainPlayListScrollPane() {
		return getPanelControlled().getPlayListTableScroll();
	}

	/**
	 * Gets the main play list table.
	 * 
	 * @return the main play list table
	 */
	public PlayListTable getMainPlayListTable() {
		return getPanelControlled().getPlayListTable();
	}

	/**
	 * Gets the selected audio objects.
	 * 
	 * @return the selected audio objects
	 */
	public List<AudioObject> getSelectedAudioObjects() {
		return getSelectedAudioObjects(VisualHandler.getInstance().getPlayListTable());
	}

	/**
	 * Gets the selected audio objects.
	 * 
	 * @param table
	 *            the table
	 * 
	 * @return the selected audio objects
	 */
	private List<AudioObject> getSelectedAudioObjects(JTable table) {
		int[] selectedRows = table.getSelectedRows();
		List<AudioObject> songs = new ArrayList<AudioObject>();
		for (int element : selectedRows) {
			AudioObject file = ((PlayListTableModel) table.getModel()).getAudioObjectAt(element);
			songs.add(file);
		}
		return songs;
	}

	/**
	 * Move down.
	 */
	public void moveDown() {
		getLogger().debug(LogCategories.CONTROLLER);

		PlayListTable table = VisualHandler.getInstance().getPlayListTable();
		int[] rows = table.getSelectedRows();
		if (rows.length > 0 && rows[rows.length - 1] < table.getRowCount() - 1) {
			((PlayListTableModel) table.getModel()).moveDown(rows);
			PlayListHandler.getInstance().moveDown(rows);
			getPanelControlled().getPlayListTable().getSelectionModel().setSelectionInterval(rows[0] + 1, rows[rows.length - 1] + 1);
		}
	}

	/**
	 * Move to bottom.
	 */
	public void moveToBottom() {
		getLogger().debug(LogCategories.CONTROLLER);

		PlayListTable table = VisualHandler.getInstance().getPlayListTable();
		int[] rows = table.getSelectedRows();
		if (rows.length > 0 && rows[rows.length - 1] < table.getRowCount() - 1) {
			((PlayListTableModel) table.getModel()).moveToBottom(rows);
			PlayListHandler.getInstance().moveToBottom(rows);
			getPanelControlled().getPlayListTable().getSelectionModel().setSelectionInterval(getPanelControlled().getPlayListTable().getRowCount() - rows.length,
					getPanelControlled().getPlayListTable().getRowCount() - 1);
		}
	}

	/**
	 * Move to top.
	 */
	public void moveToTop() {
		getLogger().debug(LogCategories.CONTROLLER);

		PlayListTable table = VisualHandler.getInstance().getPlayListTable();
		int[] rows = table.getSelectedRows();
		if (rows.length > 0 && rows[0] > 0) {
			((PlayListTableModel) table.getModel()).moveToTop(rows);
			PlayListHandler.getInstance().moveToTop(rows);
			getPanelControlled().getPlayListTable().getSelectionModel().setSelectionInterval(0, rows.length - 1);
		}
	}

	/**
	 * Move up.
	 */
	public void moveUp() {
		getLogger().debug(LogCategories.CONTROLLER);

		PlayListTable table = VisualHandler.getInstance().getPlayListTable();
		int[] rows = table.getSelectedRows();
		if (rows.length > 0 && rows[0] > 0) {
			((PlayListTableModel) table.getModel()).moveUp(rows);
			PlayListHandler.getInstance().moveUp(rows);
			getPanelControlled().getPlayListTable().getSelectionModel().setSelectionInterval(rows[0] - 1, rows[rows.length - 1] - 1);
		}
	}

	/**
	 * Notify audio objects added to controller.
	 * 
	 * @param audioObjects
	 *            the audio objects
	 * @param selected
	 *            the selected
	 */
	public void notifyAudioObjectsAddedToController(List<? extends AudioObject> audioObjects, int selected) {
		int selectedAudioObject = selected;

		if (selectedAudioObject == -1) {
			selectedAudioObject = getPanelControlled().getPlayListTable().getPlayingSong();
		}

		for (int i = 0; i < audioObjects.size(); i++) {
			((PlayListTableModel) getPanelControlled().getPlayListTable().getModel()).addSong(audioObjects.get(i));
		}

		setSelectedSongOnTable(selectedAudioObject);

		Actions.getAction(SavePlayListAction.class).setEnabled(!audioObjects.isEmpty());
		Actions.getAction(ShufflePlayListAction.class).setEnabled(!audioObjects.isEmpty());
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * net.sourceforge.atunes.kernel.controllers.model.Controller#notifyReload()
	 */
	@Override
	protected void notifyReload() {
		// Nothing to do
	}

	/**
	 * Play selected audio object.
	 */
	public void playSelectedAudioObject() {
		int audioObject = getPanelControlled().getPlayListTable().getSelectedRow();
		PlayerHandler.getInstance().setPlayListPositionToPlay(audioObject);
		PlayerHandler.getInstance().playCurrentAudioObject(false);
	}

	/**
	 * Scrolls to songs currently playing.
	 */
	public void scrollPlayList(boolean isUserAction) {
		scrollPlayList(getPanelControlled().getPlayListTable().getPlayingSong(), isUserAction);
	}

	/**
	 * Scrolls the playlist to the playing song.
	 * 
	 * @param audioObject
	 *            Audio object which should have focus
	 */
	private synchronized void scrollPlayList(int audioObject, boolean isUserAction) {
		if (!Kernel.getInstance().state.isAutoScrollPlayListEnabled() && !isUserAction) {
			return;
		}

		getLogger().debug(LogCategories.CONTROLLER, "Scrolling PlayList");

		// Get visible rectangle
		visibleRect = (Rectangle) getPanelControlled().getPlayListTable().getVisibleRect().clone();

		// Get cell height
		int heightOfRow = getPanelControlled().getPlayListTable().getCellRect(audioObject, 0, true).height;

		// Do calculation
		if (visibleRect.height == 0) {
			// At the first scroll of play list visibleRect.height is 0
			// So selected audio object is not visible
			visibleRect.y = (audioObject + 1) * heightOfRow;
		} else {
			int newYPosition = (audioObject + 1) * heightOfRow - visibleRect.height / 2;
			visibleRect.y = newYPosition;
		}

		// Correct negative value of y
		visibleRect.y = visibleRect.y >= 0 ? visibleRect.y : 0;

		// Apply scroll
		// We add a delay in order to reduce freezes
		Timer t = new Timer(1500, new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				getPanelControlled().getPlayListTable().scrollRectToVisible(visibleRect);
			}
		});
		t.setRepeats(false);
		t.start();
	}

	/**
	 * Returns panel controlled
	 */
	@Override
	protected PlayListPanel getPanelControlled() {
		return super.getPanelControlled();
	}

	/**
	 * Sets the genre.
	 */
	public void setGenre() {
		PlayListTable table = VisualHandler.getInstance().getPlayListTable();
		int[] rows = table.getSelectedRows();
		List<AudioFile> files = new ArrayList<AudioFile>();
		for (int element : rows) {
			AudioObject ao = ((PlayListTableModel) table.getModel()).getAudioObjectAt(element);
			if (ao instanceof AudioFile) {
				files.add((AudioFile) ao);
			}
		}
		TagEditionOperations.editGenre(files);
	}

	/**
	 * Sets the cover
	 */
	public void setCover() {
		PlayListTable table = VisualHandler.getInstance().getPlayListTable();
		int[] rows = table.getSelectedRows();
		List<AudioFile> files = new ArrayList<AudioFile>();
		for (int element : rows) {
			AudioObject ao = ((PlayListTableModel) table.getModel()).getAudioObjectAt(element);
			if (ao instanceof AudioFile) {
				files.add((AudioFile) ao);
			}
		}
		TagEditionOperations.editCover(files);
	}

	/**
	 * Sets the lyrics.
	 */
	public void setLyrics() {
		PlayListTable table = VisualHandler.getInstance().getPlayListTable();
		int[] rows = table.getSelectedRows();
		List<AudioFile> files = new ArrayList<AudioFile>();
		for (int element : rows) {
			AudioObject ao = ((PlayListTableModel) table.getModel()).getAudioObjectAt(element);
			if (ao instanceof AudioFile) {
				files.add((AudioFile) ao);
			}
		}
		TagEditionOperations.addLyrics(files);
	}

	/**
	 * Sets the selected song.
	 * 
	 * @param song
	 *            the new selected song
	 */
	public void setSelectedSong(int song) {
		setSelectedSongOnTable(song);
	}

	/**
	 * Sets the selected song on table.
	 * 
	 * @param song
	 *            the new selected song on table
	 */
	private void setSelectedSongOnTable(int song) {
		getPanelControlled().getPlayListTable().setPlayingSong(song);

		// Scroll automatically play list to song
		scrollPlayList(song, false);
	}

	/**
	 * Gets selected songs on play list and sets title based on artist, album
	 * and track.
	 */
	public void setTitle() {
		PlayListTable table = VisualHandler.getInstance().getPlayListTable();
		int[] rows = table.getSelectedRows();
		List<AudioFile> files = new ArrayList<AudioFile>();
		for (int element : rows) {
			AudioObject ao = ((PlayListTableModel) table.getModel()).getAudioObjectAt(element);
			if (ao instanceof AudioFile) {
				files.add((AudioFile) ao);
			}
		}
		BackgroundExecutor.changeTitles(files);
	}

	/**
	 * Sets the track number.
	 */
	public void setTrackNumber() {
		PlayListTable table = VisualHandler.getInstance().getPlayListTable();
		int[] rows = table.getSelectedRows();
		List<AudioFile> files = new ArrayList<AudioFile>();
		for (int element : rows) {
			AudioObject ao = ((PlayListTableModel) table.getModel()).getAudioObjectAt(element);
			if (ao instanceof AudioFile) {
				files.add((AudioFile) ao);
			}
		}
		TagEditionOperations.editTrackNumber(files);
	}

	/**
	 * Update position in table.
	 * 
	 * @param pos
	 *            the pos
	 */
	public void updatePositionInTable(int pos) {
		((PlayListTableModel) getPanelControlled().getPlayListTable().getModel()).refresh(pos);
	}

	/**
	 * Opens folder containing selected audio file.
	 */
	public void openFolder() {
		PlayListTable table = VisualHandler.getInstance().getPlayListTable();
		int[] rows = table.getSelectedRows();
		if (rows.length == 1) {
			AudioObject ao = ((PlayListTableModel) table.getModel()).getAudioObjectAt(rows[0]);
			if (ao instanceof AudioFile) {
				File file = ((AudioFile) ao).getParentFile();
				DesktopHandler.getInstance().openFile(file);
			}
		}
	}

	public PlayListControlsPanel getPlayListControlsPanel() {
		return getPanelControlled().getPlayListControls();
	}
}
